"""
https://github.github.com/gfm/#links
"""
import pytest

from .utils import act_and_assert


# pylint: disable=too-many-lines
@pytest.mark.gfm
def test_inline_links_493x():
    """
    Test case 493:  Here is a simple inline link:
    """

    # Arrange
    source_markdown = """[link](/uri "title")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/uri:title::::link:False:":: :]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri" title="title">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_493a():
    """
    Test case 493:  Here is a simple inline link:
    """

    # Arrange
    source_markdown = """[link](/uri "title") """
    expected_tokens = [
        "[para(1,1):: ]",
        '[link(1,1):inline:/uri:title::::link:False:":: :]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri" title="title">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_493b():
    """
    Test case 493:  Here is a simple inline link:
    """

    # Arrange
    source_markdown = """[link](/uri "title") abc"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/uri:title::::link:False:":: :]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[text(1,21): abc:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri" title="title">link</a> abc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_494():
    """
    Test case 494:  The title may be omitted:
    """

    # Arrange
    source_markdown = """[link](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri:::::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_495():
    """
    Test case 495:  (part 1) Both the title and the destination may be omitted:
    """

    # Arrange
    source_markdown = """[link]()"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline::::::link:::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_496():
    """
    Test case 496:  (part 2) Both the title and the destination may be omitted:
    """

    # Arrange
    source_markdown = """[link](<>)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline::::::link:True::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_497():
    """
    Test case 497:  (part 1) The destination can only contain spaces if it is enclosed in pointy brackets
    """

    # Arrange
    source_markdown = """[link](/my uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(/my uri):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](/my uri)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_497a():
    """
    Test case 497a:  (part 1) The destination can only contain spaces if it is enclosed in pointy brackets
    """

    # Arrange
    source_markdown = """[link](
        
[link]: /url 'title'"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):shortcut:/url:title::::link:None::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[text(1,7):(:]",
        "[end-para:::True]",
        "[BLANK(2,1):        ]",
        "[link-ref-def(3,1):True::link:: :/url:: :title:'title':]",
    ]
    expected_gfm = """<p><a href="/url" title="title">link</a>(</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_498():
    """
    Test case 498:  (part 2) The destination can only contain spaces if it is enclosed in pointy brackets
    """

    # Arrange
    source_markdown = """[link](</my uri>)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/my%20uri::/my uri:::link:True::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/my%20uri">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_499():
    """
    Test case 499:  (part 1) The destination cannot contain line breaks, even if enclosed in pointy brackets:
    """

    # Arrange
    source_markdown = """[link](foo
bar)"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(foo\nbar)::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](foo
bar)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_500():
    """
    Test case 500:  (part 2) The destination cannot contain line breaks, even if enclosed in pointy brackets:
    """

    # Arrange
    source_markdown = """[link](<foo
bar>)"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(:]",
        "[raw-html(1,8):foo\nbar]",
        "[text(2,5):):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](<foo
bar>)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_501():
    """
    Test case 501:  The destination can contain ) if it is enclosed in pointy brackets:
    """

    # Arrange
    source_markdown = """[a](<b)c>)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:b)c:::::a:True::::]",
        "[text(1,2):a:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="b)c">a</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_502():
    """
    Test case 502:  Pointy brackets that enclose links must be unescaped:
    """

    # Arrange
    source_markdown = """[link](<foo\\>)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(\a<\a&lt;\afoo\\\b\a>\a&gt;\a):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](&lt;foo&gt;)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_503():
    """
    Test case 503:  These are not links, because the opening pointy bracket is not matched properly:
    """

    # Arrange
    source_markdown = """[a](<b)c
[a](<b)c>
[a](<b>c)"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):[:]",
        "[text(1,2):a:]",
        "[text(1,3):]:]",
        "[text(1,4):(\a<\a&lt;\ab)c\n::\n]",
        "[text(2,1):[:]",
        "[text(2,2):a:]",
        "[text(2,3):]:]",
        "[text(2,4):(\a<\a&lt;\ab)c\a>\a&gt;\a\n::\n]",
        "[text(3,1):[:]",
        "[text(3,2):a:]",
        "[text(3,3):]:]",
        "[text(3,4):(:]",
        "[raw-html(3,5):b]",
        "[text(3,8):c):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[a](&lt;b)c
[a](&lt;b)c&gt;
[a](<b>c)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_504():
    """
    Test case 504:  Parentheses inside the link destination may be escaped:
    """

    # Arrange
    source_markdown = """[link](\\(foo\\))"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:(foo)::\\(foo\\):::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="(foo)">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_505():
    """
    Test case 505:  Any number of parentheses are allowed without escaping, as long as they are balanced:
    """

    # Arrange
    source_markdown = """[link](foo(and(bar)))"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:foo(and(bar)):::::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="foo(and(bar))">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_506():
    """
    Test case 506:  (part 1) However, if you have unbalanced parentheses, you need to escape or use the <...> form:
    """

    # Arrange
    source_markdown = """[link](foo\\(and\\(bar\\))"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:foo(and(bar)::foo\\(and\\(bar\\):::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="foo(and(bar)">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_507():
    """
    Test case 507:  (part 2) However, if you have unbalanced parentheses, you need to escape or use the <...> form:
    """

    # Arrange
    source_markdown = """[link](<foo(and(bar)>)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:foo(and(bar):::::link:True::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="foo(and(bar)">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_507b():
    """
    Test case 507b:  Modification of 507 without the angle brackets.
    """

    # Arrange
    source_markdown = """[link](foo(and(bar))"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(foo(and(bar)):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](foo(and(bar))</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_507c():
    """
    Test case 507c:  Modification of 507 without the angle brackets, and more open
    """

    # Arrange
    source_markdown = """[link](foo(and(b(ar))"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(foo(and(b(ar)):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](foo(and(b(ar))</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_508():
    """
    Test case 508:  Parentheses and other symbols can also be escaped, as usual in Markdown:
    """

    # Arrange
    source_markdown = """[link](foo\\)\\:)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:foo):::foo\\)\\::::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="foo):">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_509():
    """
    Test case 509:  A link can contain fragment identifiers and queries:
    """

    # Arrange
    source_markdown = """[link](#fragment)

[link](http://example.com#fragment)

[link](http://example.com?foo=3#frag)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:#fragment:::::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(2,1):]",
        "[para(3,1):]",
        "[link(3,1):inline:http://example.com#fragment:::::link:False::::]",
        "[text(3,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
        "[BLANK(4,1):]",
        "[para(5,1):]",
        "[link(5,1):inline:http://example.com?foo=3#frag:::::link:False::::]",
        "[text(5,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="#fragment">link</a></p>
<p><a href="http://example.com#fragment">link</a></p>
<p><a href="http://example.com?foo=3#frag">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_510():
    """
    Test case 510:  Note that a backslash before a non-escapable character is just a backslash:
    """

    # Arrange
    source_markdown = """[link](foo\\bar)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:foo%5Cbar::foo\\bar:::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="foo%5Cbar">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_511():
    """
    Test case 511:  URL-escaping should be left alone inside the destination
    """

    # Arrange
    source_markdown = """[link](foo%20b&auml;)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:foo%20b%C3%A4::foo%20b&auml;:::link:False::::]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="foo%20b%C3%A4">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_512():
    """
    Test case 512:  Note that, because titles can often be parsed as destinations, if you try to omit the destination and keep the title, you’ll get unexpected results:
    """

    # Arrange
    source_markdown = """[link]("title")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:%22title%22::"title":::link:False::::]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="%22title%22">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_513():
    """
    Test case 513:  Titles may be in single quotes, double quotes, or parentheses:
    """

    # Arrange
    source_markdown = """[link](/url "title")
[link](/url 'title')
[link](/url (title))"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        '[link(1,1):inline:/url:title::::link:False:":: :]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[text(1,21):\n::\n]",
        "[link(2,1):inline:/url:title::::link:False:':: :]",
        "[text(2,2):link:]",
        "[end-link::]",
        "[text(2,21):\n::\n]",
        "[link(3,1):inline:/url:title::::link:False:(:: :]",
        "[text(3,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="title">link</a>
<a href="/url" title="title">link</a>
<a href="/url" title="title">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_514():
    """
    Test case 514:  Backslash escapes and entity and numeric character references may be used in titles:
    """

    # Arrange
    source_markdown = """[link](/url "title \\"&quot;")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/url:title &quot;&quot;::title \\"&quot;::link:False:":: :]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="title &quot;&quot;">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515():
    """
    Test case 515:  Titles must be separated from the link using a whitespace. Other Unicode whitespace like non-breaking space doesn’t work.
    """

    # Arrange
    source_markdown = """[link](/url\u00A0"title")"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/url%C2%A0%22title%22::/url\u00A0"title":::link:False::::]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url%C2%A0%22title%22">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515b():
    """
    Test case 515b:  Modification of 515 to use normal space and remove closing quotes.
    """

    # Arrange
    source_markdown = """[link](/url "title)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        '[text(1,7):(/url \a"\a&quot;\atitle):]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](/url &quot;title)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515c():
    """
    Test case 515c:  Modification of 515 to use normal space and remove closing quotes.
    """

    # Arrange
    source_markdown = """[link](/url (title))"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/url:title::::link:False:(:: :]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="title">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515d():
    """
    Test case 515d:  Modification of 515 to use normal space and remove closing quotes.
    """

    # Arrange
    source_markdown = """[link](/url (title)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(/url (title):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](/url (title)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515e():
    """
    Test case 515e:  Modification of 515 to use normal space and remove closing quotes.
    """

    # Arrange
    source_markdown = """[link](/url (title(other)line))"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/url:title(other)line::::link:False:(:: :]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="title(other)line">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515f():
    """
    Test case 515e:  Modification of 515 to use normal space and remove closing quotes.
    """

    # Arrange
    source_markdown = """[link](/url (title(other)line)) abc"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/url:title(other)line::::link:False:(:: :]",
        "[text(1,2):link:]",
        "[end-link::]",
        "[text(1,32): abc:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/url" title="title(other)line">link</a> abc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_515g():
    """
    Test case 515e:  Modification of 515 to use normal space and remove closing quotes.
    """

    # Arrange
    source_markdown = """[link](/url (title(otherline) abc"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7):(/url (title(otherline) abc:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](/url (title(otherline) abc</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_516():
    """
    Test case 516:  Nested balanced quotes are not allowed without escaping:
    """

    # Arrange
    source_markdown = """[link](/url "title "and" title")"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        '[text(1,7):(/url \a"\a&quot;\atitle \a"\a&quot;\aand\a"\a&quot;\a title\a"\a&quot;\a):]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link](/url &quot;title &quot;and&quot; title&quot;)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_517():
    """
    Test case 517:  But it is easy to work around this by using a different quote type:
    """

    # Arrange
    source_markdown = """[link](/url 'title "and" title')"""
    expected_tokens = [
        "[para(1,1):]",
        '[link(1,1):inline:/url:title &quot;and&quot; title::title "and" title::link:False:\':: :]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p><a href="/url" title="title &quot;and&quot; title">link</a></p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_518():
    """
    Test case 518:  Whitespace is allowed around the destination and title:
    """

    # Arrange
    source_markdown = """[link](   /uri
  "title"  )"""
    expected_tokens = [
        "[para(1,1):\n  ]",
        '[link(1,1):inline:/uri:title::::link:False:":   :\n:  ]',
        "[text(1,2):link:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri" title="title">link</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_519():
    """
    Test case 519:  But it is not allowed between the link text and the following parenthesis:
    """

    # Arrange
    source_markdown = """[link] (/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7): (/uri):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link] (/uri)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_520():
    """
    Test case 520:  (part 1) The link text may contain balanced brackets, but not unbalanced ones, unless they are escaped:
    """

    # Arrange
    source_markdown = """[link [foo [bar]]](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri:::::link [foo [bar]]:False::::]",
        "[text(1,2):link :]",
        "[text(1,7):[:]",
        "[text(1,8):foo :]",
        "[text(1,12):[:]",
        "[text(1,13):bar:]",
        "[text(1,16):]:]",
        "[text(1,17):]:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">link [foo [bar]]</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_521():
    """
    Test case 521:  (part 2) The link text may contain balanced brackets, but not unbalanced ones, unless they are escaped:
    """

    # Arrange
    source_markdown = """[link] bar](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link:]",
        "[text(1,6):]:]",
        "[text(1,7): bar:]",
        "[text(1,11):]:]",
        "[text(1,12):(/uri):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link] bar](/uri)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_522():
    """
    Test case 522:  (part 3) The link text may contain balanced brackets, but not unbalanced ones, unless they are escaped:
    """

    # Arrange
    source_markdown = """[link [bar](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):link :]",
        "[link(1,7):inline:/uri:::::bar:False::::]",
        "[text(1,8):bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[link <a href="/uri">bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_523():
    """
    Test case 523:  (part 4) The link text may contain balanced brackets, but not unbalanced ones, unless they are escaped:
    """

    # Arrange
    source_markdown = """[link \\[bar](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri:::::link \\[bar:False::::]",
        "[text(1,2):link \\\b[bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">link [bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_524():
    """
    Test case 524:  (part 1) The link text may contain inline content:
    """

    # Arrange
    source_markdown = """[link *foo **bar** `#`*](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri:::::link *foo **bar** `#`*:False::::]",
        "[text(1,2):link :]",
        "[emphasis(1,7):1:*]",
        "[text(1,8):foo :]",
        "[emphasis(1,12):2:*]",
        "[text(1,14):bar:]",
        "[end-emphasis(1,17)::]",
        "[text(1,19): :]",
        "[icode-span(1,20):#:`::]",
        "[end-emphasis(1,23)::]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri">link <em>foo <strong>bar</strong> <code>#</code></em></a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_525():
    """
    Test case 525:  (part 2) The link text may contain inline content:
    """

    # Arrange
    source_markdown = """[![moon](moon.jpg)](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:/uri:::::![moon](moon.jpg):False::::]",
        "[image(1,2):inline:moon.jpg::moon::::moon:False::::]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="/uri"><img src="moon.jpg" alt="moon" /></a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_526():
    """
    Test case 526:  (part 1) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """[foo [bar](/uri)](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        "[link(1,6):inline:/uri:::::bar:False::::]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[text(1,17):]:]",
        "[text(1,18):(/uri):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo <a href="/uri">bar</a>](/uri)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_526a():
    """
    Test case 526a:  (part 1) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """[foo [bar](/uri1)](/uri2)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        "[link(1,6):inline:/uri1:::::bar:False::::]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[text(1,18):]:]",
        "[text(1,19):(/uri2):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo <a href="/uri1">bar</a>](/uri2)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_527():
    """
    Test case 527:  (part 2) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """[foo *[bar [baz](/uri)](/uri)*](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):1:*]",
        "[text(1,7):[:]",
        "[text(1,8):bar :]",
        "[link(1,12):inline:/uri:::::baz:False::::]",
        "[text(1,13):baz:]",
        "[end-link::]",
        "[text(1,23):]:]",
        "[text(1,24):(/uri):]",
        "[end-emphasis(1,30)::]",
        "[text(1,31):]:]",
        "[text(1,32):(/uri):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo <em>[bar <a href="/uri">baz</a>](/uri)</em>](/uri)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_527a():
    """
    Test case 527a:  (part 2) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """[foo *[bar [baz](/uri1)](/uri2)*](/uri3)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        "[emphasis(1,6):1:*]",
        "[text(1,7):[:]",
        "[text(1,8):bar :]",
        "[link(1,12):inline:/uri1:::::baz:False::::]",
        "[text(1,13):baz:]",
        "[end-link::]",
        "[text(1,24):]:]",
        "[text(1,25):(/uri2):]",
        "[end-emphasis(1,32)::]",
        "[text(1,33):]:]",
        "[text(1,34):(/uri3):]",
        "[end-para:::True]",
    ]
    expected_gfm = (
        """<p>[foo <em>[bar <a href="/uri1">baz</a>](/uri2)</em>](/uri3)</p>"""
    )

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_528():
    """
    Test case 528:  (part 3) However, links may not contain other links, at any level of nesting.
    """

    # Arrange
    source_markdown = """![[[foo](uri1)](uri2)](uri3)"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:uri3::[foo](uri2)::::[[foo](uri1)](uri2):False::::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><img src="uri3" alt="[foo](uri2)" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_528a():
    """
    Test case 528a:  variation
    """

    # Arrange
    source_markdown = """![[foo](uri2)](uri3)"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:uri3::foo::::[foo](uri2):False::::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><img src="uri3" alt="foo" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_528b():
    """
    Test case 528b:  variation
    """

    # Arrange
    source_markdown = """![[foo](uri2 "bar")](uri3)"""
    expected_tokens = [
        "[para(1,1):]",
        '[image(1,1):inline:uri3::foo::::[foo](uri2 "bar"):False::::]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p><img src="uri3" alt="foo" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_528c():
    """
    Test case 528c:  variation
    """

    # Arrange
    source_markdown = """![![[foo](uri1)](uri2)](uri3)"""
    expected_tokens = [
        "[para(1,1):]",
        "[image(1,1):inline:uri3::foo::::![[foo](uri1)](uri2):False::::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><img src="uri3" alt="foo" /></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_529():
    """
    Test case 529:  (part 1) These cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """*[foo*](/uri)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[link(1,2):inline:/uri:::::foo*:False::::]",
        "[text(1,3):foo:]",
        "[text(1,6):*:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*<a href="/uri">foo*</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_530():
    """
    Test case 530:  (part 2) These cases illustrate the precedence of link text grouping over emphasis grouping:
    """

    # Arrange
    source_markdown = """[foo *bar](baz*)"""
    expected_tokens = [
        "[para(1,1):]",
        "[link(1,1):inline:baz*:::::foo *bar:False::::]",
        "[text(1,2):foo :]",
        "[text(1,6):*:]",
        "[text(1,7):bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="baz*">foo *bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_531():
    """
    Test case 531:  Note that brackets that aren’t part of links do not take precedence:
    """

    # Arrange
    source_markdown = """*foo [bar* baz]"""
    expected_tokens = [
        "[para(1,1):]",
        "[emphasis(1,1):1:*]",
        "[text(1,2):foo :]",
        "[text(1,6):[:]",
        "[text(1,7):bar:]",
        "[end-emphasis(1,10)::]",
        "[text(1,11): baz:]",
        "[text(1,15):]:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><em>foo [bar</em> baz]</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_532():
    """
    Test case 532:  (part 1) These cases illustrate the precedence of HTML tags, code spans, and autolinks over link grouping:
    """

    # Arrange
    source_markdown = """[foo <bar attr="](baz)">"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo :]",
        '[raw-html(1,6):bar attr="](baz)"]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo <bar attr="](baz)"></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_533():
    """
    Test case 533:  (part 2) These cases illustrate the precedence of HTML tags, code spans, and autolinks over link grouping:
    """

    # Arrange
    source_markdown = """[foo`](/uri)`"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[icode-span(1,5):](/uri):`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo<code>](/uri)</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_534():
    """
    Test case 534:  (part 3) These cases illustrate the precedence of HTML tags, code spans, and autolinks over link grouping:
    """

    # Arrange
    source_markdown = """[foo<http://example.com/?search=](uri)>"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):foo:]",
        "[uri-autolink(1,5):http://example.com/?search=](uri)]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[foo<a href="http://example.com/?search=%5D(uri)">http://example.com/?search=](uri)</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_extra_01():
    """
    Test case extra 01:  variation on 644
    """

    # Arrange
    source_markdown = """[foo <!-- this is a
comment - with hyphen --> bar](uri)"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[link(1,1):inline:uri:::::foo <!-- this is a\ncomment - with hyphen --> bar:False::::]",
        "[text(1,2):foo :]",
        "[raw-html(1,6):!-- this is a\ncomment - with hyphen --]",
        "[text(2,26): bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="uri">foo <!-- this is a
comment - with hyphen --> bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_inline_links_extra_02():
    """
    Test case extra 02:  variation on 345
    """

    # Arrange
    source_markdown = """[foo ``
foo
bar\a\a
baz
`` bar](uri)""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n\n\n\n]",
        "[link(1,1):inline:uri:::::foo ``\nfoo\nbar  \nbaz\n`` bar:False::::]",
        "[text(1,2):foo :]",
        "[icode-span(1,6):foo\a\n\a \abar  \a\n\a \abaz:``:\a\n\a \a:\a\n\a \a]",
        "[text(5,3): bar:]",
        "[end-link::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="uri">foo <code>foo bar   baz</code> bar</a></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Inline helper
"""
import json
import logging
import os
import re
import string

from pymarkdown.bad_tokenization_error import BadTokenizationError
from pymarkdown.html_helper import HtmlHelper
from pymarkdown.inline_markdown_token import (
    EmailAutolinkMarkdownToken,
    HardBreakMarkdownToken,
    InlineCodeSpanMarkdownToken,
    UriAutolinkMarkdownToken,
)
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods, too-many-instance-attributes
class InlineRequest:
    """
    Class to hold the request information to pass on to the handle_* functions.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        source_text,
        next_index,
        inline_blocks=None,
        remaining_line=None,
        current_string_unresolved=None,
        line_number=None,
        column_number=None,
        para_owner=None,
    ):
        (
            self.source_text,
            self.next_index,
            self.inline_blocks,
            self.remaining_line,
            self.current_string_unresolved,
            self.line_number,
            self.column_number,
            self.para_owner,
        ) = (
            source_text,
            next_index,
            inline_blocks,
            remaining_line,
            current_string_unresolved,
            line_number,
            column_number,
            para_owner,
        )

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods, too-many-instance-attributes


# pylint: disable=too-few-public-methods,too-many-arguments,too-many-instance-attributes
class InlineResponse:
    """
    Class to hold the response from the inline handle_* functions.
    """

    def __init__(self):
        (
            self.new_string,
            self.new_index,
            self.new_tokens,
            self.new_string_unresolved,
            self.consume_rest_of_line,
            self.original_string,
            self.delta_line_number,
            self.delta_column_number,
        ) = (None, None, None, None, False, None, 0, 0)
        self.clear_fields()

    def clear_fields(self):
        """
        Clear any of the fields that start with new_*.
        """
        (
            self.new_string,
            self.new_index,
            self.new_tokens,
            self.new_string_unresolved,
            self.consume_rest_of_line,
            self.original_string,
            self.delta_line_number,
            self.delta_column_number,
        ) = (None, None, None, None, False, None, 0, 0)


# pylint: enable=too-few-public-methods,too-many-arguments,too-many-instance-attributes


class InlineHelper:
    """
    Class to helper with the parsing of inline elements.
    """

    __valid_email_regex = "^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$"

    __scheme_end_character = ":"
    __valid_scheme_characters = f"{string.ascii_letters}{string.digits}.-+"

    __backslash_punctuation = "!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~\\"
    __html_character_escape_map = {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
        '"': "&quot;",
    }

    angle_bracket_start = "<"
    __angle_bracket_end = ">"
    code_span_bounds = "`"
    backslash_character = "\\"
    __entity_map = {}
    character_reference_start_character = "&"
    __numeric_character_reference_start_character = "#"
    __hex_character_reference_start_character = "xX"
    __character_reference_end_character = ";"
    __invalid_reference_character_substitute = "\ufffd"
    __line_end_whitespace = ParserHelper.space_character
    __valid_backslash_sequence_starts = (
        f"{backslash_character}{character_reference_start_character}"
    )

    __ascii_letters_and_digits = f"{string.ascii_letters}{string.digits}"

    __skip_html5_entities_ending_with = ";"

    __entities_file_name = "entities.json"

    @staticmethod
    def initialize(resource_path):
        """
        Initialize the inline subsystem.
        """
        InlineHelper.__entity_map = InlineHelper.__load_entity_map(resource_path)

    @staticmethod
    def handle_inline_backslash(inline_request, add_text_signature=True):
        """
        Handle the inline case of having a backslash.
        """

        inline_response = InlineResponse()
        (
            inline_response.new_index,
            inline_response.new_string,
            inline_response.new_string_unresolved,
            source_text_size,
        ) = (inline_request.next_index + 1, "", "", len(inline_request.source_text))
        if inline_response.new_index >= source_text_size or (
            inline_response.new_index < source_text_size
            and inline_request.source_text[inline_response.new_index]
            == ParserHelper.newline_character
        ):
            inline_response.new_string, inline_response.new_string_unresolved = (
                InlineHelper.backslash_character,
                inline_response.new_string,
            )
        else:
            if (
                inline_request.source_text[inline_response.new_index]
                in InlineHelper.__backslash_punctuation
            ):
                inline_response.new_string = (
                    ParserHelper.backslash_escape_sequence if add_text_signature else ""
                ) + inline_request.source_text[inline_response.new_index]
                inline_response.new_string_unresolved = (
                    InlineHelper.backslash_character + inline_response.new_string
                )
            else:
                inline_response.new_string = (
                    InlineHelper.backslash_character
                    + inline_request.source_text[inline_response.new_index]
                )
                inline_response.new_string_unresolved = inline_response.new_string
            inline_response.new_index += 1

        inline_response.delta_line_number, inline_response.delta_column_number = (
            0,
            inline_response.new_index - inline_request.next_index,
        )
        return inline_response

    @staticmethod
    def handle_character_reference(inline_request):
        """
        Handle a generic character reference.
        """
        inline_response = InlineResponse()
        inline_response.new_index, inline_response.new_string, source_text_size = (
            inline_request.next_index + 1,
            "",
            len(inline_request.source_text),
        )
        if (
            inline_response.new_index < source_text_size
            and inline_request.source_text[inline_response.new_index]
            == InlineHelper.__numeric_character_reference_start_character
        ):
            original_new_index = inline_response.new_index
            POGGER.debug("here")
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.original_string,
            ) = InlineHelper.__handle_numeric_character_reference(
                inline_request.source_text, inline_response.new_index
            )
            inline_response.new_string_unresolved = (
                InlineHelper.character_reference_start_character
                + inline_request.source_text[
                    original_new_index : inline_response.new_index
                ]
            )
            POGGER.debug("here-->$<--", inline_response.new_string)
            POGGER.debug("here-->$<--", inline_response.new_string_unresolved)
        else:
            POGGER.debug("there")
            end_index, collected_string = ParserHelper.collect_while_one_of_characters(
                inline_request.source_text,
                inline_response.new_index,
                InlineHelper.__ascii_letters_and_digits,
            )
            if collected_string:
                collected_string = (
                    InlineHelper.character_reference_start_character + collected_string
                )
                if (
                    end_index < source_text_size
                    and inline_request.source_text[end_index]
                    == InlineHelper.__character_reference_end_character
                ):
                    end_index += 1
                    collected_string += InlineHelper.__character_reference_end_character
                    original_collected_string = collected_string
                    if collected_string in InlineHelper.__entity_map:
                        inline_response.original_string = collected_string
                        collected_string = InlineHelper.__entity_map[collected_string]
                        inline_response.new_string_unresolved = (
                            original_collected_string
                        )
                inline_response.new_string, inline_response.new_index = (
                    collected_string,
                    end_index,
                )
                POGGER.debug("there-->$<--", inline_response.new_string)
                POGGER.debug("there-->$<--", inline_response.new_string_unresolved)
            else:
                inline_response.new_string = (
                    InlineHelper.character_reference_start_character
                )
        inline_response.delta_line_number, inline_response.delta_column_number = (
            0,
            inline_response.new_index - inline_request.next_index,
        )
        return inline_response

    @staticmethod
    def handle_backslashes(source_text, add_text_signature=True):
        """
        Handle the processing of backslashes for anything other than the text
        blocks, which have additional needs for parsing.
        """

        start_index = 0
        next_index = ParserHelper.index_any_of(
            source_text, InlineHelper.__valid_backslash_sequence_starts, start_index
        )
        string_parts = []
        while next_index != -1:
            string_parts.append(source_text[start_index:next_index])
            current_char = source_text[next_index]

            inline_request = InlineRequest(source_text, next_index)
            POGGER.debug("handle_backslashes>>$>>", current_char)
            if current_char == InlineHelper.backslash_character:
                inline_response = InlineHelper.handle_inline_backslash(
                    inline_request, add_text_signature=add_text_signature
                )
                new_string, new_index = (
                    inline_response.new_string,
                    inline_response.new_index,
                )
            else:
                assert (
                    source_text[next_index]
                    == InlineHelper.character_reference_start_character
                )
                inline_response = InlineHelper.handle_character_reference(
                    inline_request
                )
                new_string, new_index = (
                    inline_response.new_string,
                    inline_response.new_index,
                )

            POGGER.debug("handle_backslashes<<$<<$", new_string, new_index)
            string_parts.append(new_string)
            start_index = new_index
            next_index = ParserHelper.index_any_of(
                source_text, InlineHelper.__valid_backslash_sequence_starts, start_index
            )

        if start_index < len(source_text):
            string_parts.append(source_text[start_index:])
        return "".join(string_parts)

    @staticmethod
    def append_text(
        string_to_append_to,
        text_to_append,
        alternate_escape_map=None,
        add_text_signature=True,
    ):
        """
        Append the text to the given string, doing any needed encoding as we go.
        """

        if not alternate_escape_map:
            alternate_escape_map = InlineHelper.__html_character_escape_map

        start_index = 0
        next_index = ParserHelper.index_any_of(
            text_to_append, alternate_escape_map.keys(), start_index
        )
        text_parts = [string_to_append_to]
        while next_index != -1:

            escaped_part = alternate_escape_map[text_to_append[next_index]]
            text_parts.extend(
                [
                    text_to_append[start_index:next_index],
                    ParserHelper.create_replacement_markers(
                        text_to_append[next_index], escaped_part
                    )
                    if add_text_signature
                    else escaped_part,
                ]
            )

            start_index = next_index + 1
            next_index = ParserHelper.index_any_of(
                text_to_append, alternate_escape_map.keys(), start_index
            )

        if start_index < len(text_to_append):
            text_parts.append(text_to_append[start_index:])

        return "".join(text_parts)

    # pylint: disable=too-many-statements, too-many-locals
    @staticmethod
    def handle_inline_backtick(inline_request):
        """
        Handle the inline case of backticks for code spans.
        """
        POGGER.debug("before_collect>$", inline_request.next_index)
        (
            new_index,
            extracted_start_backticks,
        ) = ParserHelper.collect_while_one_of_characters(
            inline_request.source_text,
            inline_request.next_index,
            InlineHelper.code_span_bounds,
        )
        POGGER.debug("after_collect>$>$", new_index, extracted_start_backticks)

        extracted_start_backticks_size, end_backtick_start_index = (
            len(extracted_start_backticks),
            inline_request.source_text.find(extracted_start_backticks, new_index),
        )
        while end_backtick_start_index != -1:
            (
                end_backticks_index,
                end_backticks_attempt,
            ) = ParserHelper.collect_while_one_of_characters(
                inline_request.source_text,
                end_backtick_start_index,
                InlineHelper.code_span_bounds,
            )
            if len(end_backticks_attempt) == extracted_start_backticks_size:
                break
            end_backtick_start_index = inline_request.source_text.find(
                extracted_start_backticks, end_backticks_index
            )

        inline_response = InlineResponse()
        inline_response.delta_line_number = -1

        if end_backtick_start_index == -1:
            inline_response.new_string, inline_response.new_index = (
                extracted_start_backticks,
                new_index,
            )
        else:
            between_text = inline_request.source_text[
                new_index:end_backtick_start_index
            ]
            original_between_text = between_text
            POGGER.debug(
                "after_collect>$>>$>>$<<",
                between_text,
                end_backtick_start_index,
                inline_request.source_text[end_backtick_start_index:],
            )
            leading_whitespace, trailing_whitespace = "", ""
            if (
                len(between_text) > 2
                and (
                    between_text[0] == ParserHelper.space_character
                    or between_text[0] == ParserHelper.newline_character
                )
                and (
                    between_text[-1] == ParserHelper.space_character
                    or between_text[-1] == ParserHelper.newline_character
                )
            ):
                stripped_between_attempt = between_text[1:-1]
                if len(stripped_between_attempt.strip()) != 0:
                    leading_whitespace, trailing_whitespace = (
                        between_text[0],
                        between_text[-1],
                    )
                    between_text = stripped_between_attempt

            replaced_newline = ParserHelper.create_replacement_markers(
                ParserHelper.newline_character, ParserHelper.space_character
            )
            POGGER.debug("between_text>>$<<", between_text)
            between_text = ParserHelper.escape_special_characters(between_text)
            POGGER.debug("between_text>>$<<", between_text)
            POGGER.debug(
                "leading_whitespace>>$<<",
                leading_whitespace,
            )
            POGGER.debug(
                "trailing_whitespace>>$<<",
                trailing_whitespace,
            )
            between_text, leading_whitespace, trailing_whitespace = (
                between_text.replace(ParserHelper.newline_character, replaced_newline),
                leading_whitespace.replace(
                    ParserHelper.newline_character, replaced_newline
                ),
                trailing_whitespace.replace(
                    ParserHelper.newline_character, replaced_newline
                ),
            )

            POGGER.debug("between_text>>$<<", between_text)
            between_text = InlineHelper.append_text("", between_text)
            POGGER.debug("between_text>>$<<", between_text)
            POGGER.debug(
                "leading_whitespace>>$<<",
                leading_whitespace,
            )
            POGGER.debug(
                "trailing_whitespace>>$<<",
                trailing_whitespace,
            )
            end_backtick_start_index += extracted_start_backticks_size
            inline_response.new_string, inline_response.new_index = (
                "",
                end_backtick_start_index,
            )
            new_column_number = inline_request.column_number + len(
                inline_request.remaining_line
            )
            POGGER.debug(">>new_column_number>>$", new_column_number)

            inline_response.new_tokens = [
                InlineCodeSpanMarkdownToken(
                    between_text,
                    extracted_start_backticks,
                    leading_whitespace,
                    trailing_whitespace,
                    inline_request.line_number,
                    new_column_number,
                )
            ]

            if ParserHelper.newline_character in original_between_text:
                (
                    inline_response.delta_line_number,
                    inline_response.delta_column_number,
                ) = ParserHelper.calculate_deltas(
                    original_between_text + extracted_start_backticks
                )

        POGGER.debug(
            ">>delta_line_number>>$<<",
            inline_response.delta_line_number,
        )
        POGGER.debug(
            ">>delta_column_number>>$<<",
            inline_response.delta_column_number,
        )
        if inline_response.delta_line_number == -1:
            inline_response.delta_line_number, inline_response.delta_column_number = (
                0,
                inline_response.new_index - inline_request.next_index,
            )
        return inline_response

    # pylint: enable=too-many-statements, too-many-locals

    @staticmethod
    def modify_end_string(end_string, removed_end_whitespace):
        """
        Modify the string at the end of the paragraph.
        """
        if end_string is None:
            return f"{removed_end_whitespace}{ParserHelper.newline_character}"
        return f"{end_string}{removed_end_whitespace}{ParserHelper.newline_character}"

    # pylint: disable=too-many-arguments, too-many-locals, too-many-statements
    @staticmethod
    def handle_line_end(
        next_index,
        remaining_line,
        end_string,
        current_string,
        inline_blocks,
        is_setext,
        line_number,
        column_number,
        coalesced_stack,
    ):
        """
        Handle the inline case of having the end of line character encountered.
        """
        new_tokens = []

        POGGER.debug(">>current_string>>$>>", current_string)
        POGGER.debug(">>end_string>>$>>", end_string)
        POGGER.debug(">>remaining_line>>$>>", remaining_line)
        _, last_non_whitespace_index = ParserHelper.collect_backwards_while_character(
            remaining_line, -1, InlineHelper.__line_end_whitespace
        )
        POGGER.debug(">>last_non_whitespace_index>>$", last_non_whitespace_index)
        removed_end_whitespace = remaining_line[last_non_whitespace_index:]
        remaining_line = remaining_line[0:last_non_whitespace_index]
        POGGER.debug(">>removed_end_whitespace>>$>>", removed_end_whitespace)
        POGGER.debug(">>remaining_line>>$>>", remaining_line)

        POGGER.debug(">>current_string>>$>>", current_string)
        (
            append_to_current_string,
            whitespace_to_add,
            removed_end_whitespace_size,
            adj_hard_column,
        ) = (
            ParserHelper.newline_character,
            None,
            len(removed_end_whitespace),
            column_number + len(remaining_line),
        )
        POGGER.debug(
            ">>len(r_e_w)>>$>>rem>>$>>",
            removed_end_whitespace_size,
            remaining_line,
        )
        POGGER.debug(">>current_string>>$>>", current_string)

        is_proper_hard_break, current_string_size = False, len(current_string)
        if (
            removed_end_whitespace_size == 0
            and current_string_size
            and current_string[current_string_size - 1]
            == InlineHelper.backslash_character
        ):
            POGGER.debug(">>$<<", current_string)
            modified_current_string = current_string[0:-1]
            is_proper_hard_break = modified_current_string[-2:] != "\\\b"
            POGGER.debug(">>$<<", is_proper_hard_break)

        if is_proper_hard_break:
            POGGER.debug(">>proper hard break")
            new_tokens.append(
                HardBreakMarkdownToken(
                    InlineHelper.backslash_character, line_number, adj_hard_column - 1
                )
            )
            current_string, whitespace_to_add = current_string[0:-1], None
            append_to_current_string = ""
        elif removed_end_whitespace_size >= 2:
            POGGER.debug(">>whitespace hard break")
            new_tokens.append(
                HardBreakMarkdownToken(
                    removed_end_whitespace, line_number, adj_hard_column
                )
            )
            whitespace_to_add = None
            append_to_current_string = ""
        else:
            POGGER.debug(">>normal end")
            POGGER.debug("<<is_setext<<$<<", is_setext)
            POGGER.debug("<<inline_blocks<<$<<", inline_blocks)
            POGGER.debug("<<current_string<<$<<", current_string)
            POGGER.debug("<<remaining_line<<$<<", remaining_line)
            POGGER.debug("<<end_string<<$<<", end_string)
            POGGER.debug("<<removed_end_whitespace<<$<<", removed_end_whitespace)
            if (
                is_setext
                and inline_blocks
                and inline_blocks[-1].is_inline_hard_break
                and not current_string
            ):
                new_index, ex_ws = ParserHelper.extract_whitespace(remaining_line, 0)
                POGGER.debug("<<new_index<<$<<", new_index)
                POGGER.debug("<<ex_ws<<$<<", ex_ws)
                assert new_index
                end_string = ex_ws + ParserHelper.whitespace_split_character
                remaining_line = remaining_line[new_index:]

            end_string = InlineHelper.modify_end_string(
                end_string, removed_end_whitespace
            )
            POGGER.debug("<<end_string<<$<<", end_string)

        if coalesced_stack and coalesced_stack[-1].is_block_quote_start:
            coalesced_stack[-1].leading_text_index += 1

        POGGER.debug(
            "<<append_to_current_string<<$<<",
            append_to_current_string,
        )
        POGGER.debug(
            "<<whitespace_to_add<<$<<",
            whitespace_to_add,
        )
        POGGER.debug("<<remaining_line<<$<<", remaining_line)
        POGGER.debug("<<end_string<<$<<", end_string)
        POGGER.debug("<<current_string<<$<<", current_string)
        return (
            append_to_current_string,
            whitespace_to_add,
            next_index + 1,
            new_tokens,
            remaining_line,
            end_string,
            current_string,
        )

    # pylint: enable=too-many-arguments, too-many-locals, too-many-statements

    @staticmethod
    def extract_bounded_string(
        source_text, new_index, close_character, start_character
    ):
        """
        Extract a string that is bounded by some manner of characters.
        """
        break_characters = (
            f"{InlineHelper.backslash_character}{close_character}{start_character}"
            if start_character
            else f"{InlineHelper.backslash_character}{close_character}"
        )
        nesting_level = 0
        POGGER.debug(
            "extract_bounded_string>>new_index>>$>>data>>$>>",
            new_index,
            source_text[new_index:],
        )
        next_index, data = ParserHelper.collect_until_one_of_characters(
            source_text, new_index, break_characters
        )
        extracted_parts = [data]
        POGGER.debug(
            ">>next_index1>>$>>data>>$>>",
            next_index,
            data,
        )
        source_text_size = len(source_text)
        while next_index < source_text_size and not (
            source_text[next_index] == close_character and nesting_level == 0
        ):
            if ParserHelper.is_character_at_index(
                source_text, next_index, InlineHelper.backslash_character
            ):
                POGGER.debug("pre-back>>next_index>>$>>", next_index)
                old_index = next_index

                inline_request = InlineRequest(source_text, next_index)
                inline_response = InlineHelper.handle_inline_backslash(inline_request)
                next_index = inline_response.new_index
                extracted_parts.append(source_text[old_index:next_index])
            elif start_character is not None and ParserHelper.is_character_at_index(
                source_text, next_index, start_character
            ):
                POGGER.debug("pre-start>>next_index>>$>>", next_index)
                extracted_parts.append(start_character)
                next_index += 1
                nesting_level += 1
            else:
                assert ParserHelper.is_character_at_index(
                    source_text, next_index, close_character
                )
                POGGER.debug("pre-close>>next_index>>$>>", next_index)
                extracted_parts.append(close_character)
                next_index += 1
                nesting_level -= 1
            next_index, new_data = ParserHelper.collect_until_one_of_characters(
                source_text, next_index, break_characters
            )
            POGGER.debug(
                "back>>next_index>>$>>data>>$>>",
                next_index,
                data,
            )
            extracted_parts.append(new_data)
        POGGER.debug(
            ">>next_index2>>$>>data>>$>>",
            next_index,
            data,
        )
        if (
            ParserHelper.is_character_at_index(source_text, next_index, close_character)
            and nesting_level == 0
        ):
            POGGER.debug("extract_bounded_string>>found-close")
            return next_index + 1, "".join(extracted_parts)
        POGGER.debug(
            "extract_bounded_string>>ran out of string>>next_index>>$", next_index
        )
        return next_index, None

    @staticmethod
    def __handle_numeric_character_reference(source_text, new_index):
        """
        Handle a character reference that is numeric in nature.
        """

        original_reference, new_index, translated_reference, source_text_size = (
            None,
            new_index + 1,
            -1,
            len(source_text),
        )
        if new_index < source_text_size and (
            source_text[new_index]
            in InlineHelper.__hex_character_reference_start_character
        ):
            hex_char = source_text[new_index]
            new_index += 1
            end_index, collected_string = ParserHelper.collect_while_one_of_characters(
                source_text, new_index, string.hexdigits
            )
            POGGER.debug(
                "&#x>>a>>$>>b>>$>>$",
                end_index,
                collected_string,
                source_text_size,
            )
            delta = end_index - new_index
            POGGER.debug("delta>>$>>", delta)
            if 1 <= delta <= 6:
                translated_reference = int(collected_string, 16)
            new_string, new_index = (
                f"{InlineHelper.character_reference_start_character}{InlineHelper.__numeric_character_reference_start_character}{hex_char}{collected_string}",
                end_index,
            )
        else:
            end_index, collected_string = ParserHelper.collect_while_one_of_characters(
                source_text, new_index, string.digits
            )
            POGGER.debug(
                "&#>>a>>$>>b>>$>>$",
                end_index,
                collected_string,
                source_text_size,
            )
            delta = end_index - new_index
            POGGER.debug("delta>>$>>", delta)
            if 1 <= delta <= 7:
                translated_reference = int(collected_string)
            new_string, new_index = (
                f"{InlineHelper.character_reference_start_character}{InlineHelper.__numeric_character_reference_start_character}{collected_string}",
                end_index,
            )

        if (
            translated_reference >= 0
            and new_index < source_text_size
            and source_text[new_index]
            == InlineHelper.__character_reference_end_character
        ):
            new_index += 1
            original_reference = f"{new_string};"
            new_string = (
                InlineHelper.__invalid_reference_character_substitute
                if translated_reference == 0
                else chr(translated_reference)
            )
        return new_string, new_index, original_reference

    @staticmethod
    def __load_entity_map(resource_path):
        """
        Load the entity map, refreshed from https://html.spec.whatwg.org/entities.json
        into a dict that was can use.
        """

        master_entities_file = os.path.join(
            resource_path, InlineHelper.__entities_file_name
        )
        try:
            with open(os.path.abspath(master_entities_file)) as infile:
                results_dictionary = json.load(infile)
        except json.decoder.JSONDecodeError as this_exception:
            error_message = f"Named character entity map file '{master_entities_file}' is not a valid JSON file ({str(this_exception)})."
            raise BadTokenizationError(error_message) from this_exception
        except IOError as this_exception:
            error_message = f"Named character entity map file '{master_entities_file}' was not loaded ({str(this_exception)})."
            raise BadTokenizationError(error_message) from this_exception

        approved_entity_map = {}
        for next_name in results_dictionary:

            # Downloaded file is for HTML5, which includes some names that do
            # not end with ";".  These are excluded.
            if not next_name[-1] == InlineHelper.__skip_html5_entities_ending_with:
                continue

            char_entity = results_dictionary[next_name]
            entity_characters = char_entity["characters"]
            entity_codepoints = char_entity["codepoints"]

            # The only entities we should encounter either have a length of 1 or 2
            if len(entity_characters) == 1:
                assert len(entity_codepoints) == 1
                assert ord(entity_characters[0]) == entity_codepoints[0]
            else:
                assert len(entity_codepoints) == 2
                assert ord(entity_characters[0]) == entity_codepoints[0]
                assert ord(entity_characters[1]) == entity_codepoints[1]
            approved_entity_map[next_name] = entity_characters
        return approved_entity_map

    @staticmethod
    def __parse_valid_email_autolink(text_to_parse, line_number, column_number):
        """
        Parse a possible email autolink and determine if it is valid.
        """
        if re.match(InlineHelper.__valid_email_regex, text_to_parse):
            return EmailAutolinkMarkdownToken(text_to_parse, line_number, column_number)
        return None

    @staticmethod
    def __parse_valid_uri_autolink(text_to_parse, line_number, column_number):
        """
        Parse a possible uri autolink and determine if it is valid.
        """

        uri_scheme, path_index = "", -1
        if (
            InlineHelper.angle_bracket_start not in text_to_parse
            and text_to_parse[0] in string.ascii_letters
        ):
            path_index, uri_scheme = ParserHelper.collect_while_one_of_characters(
                text_to_parse, 1, InlineHelper.__valid_scheme_characters
            )
            uri_scheme, text_to_parse_size = f"{text_to_parse[0]}{uri_scheme}", len(
                text_to_parse
            )
            if (
                2 <= len(uri_scheme) <= 32
                and path_index < text_to_parse_size
                and text_to_parse[path_index] == InlineHelper.__scheme_end_character
            ):
                path_index += 1
                while path_index < text_to_parse_size:
                    if ord(text_to_parse[path_index]) <= 32:
                        break
                    path_index += 1
                if path_index == text_to_parse_size:
                    return UriAutolinkMarkdownToken(
                        text_to_parse, line_number, column_number
                    )
        return None

    @staticmethod
    def handle_angle_brackets(inline_request):
        """
        Given an open angle bracket, determine which of the three possibilities it is.
        """
        closing_angle_index = inline_request.source_text.find(
            InlineHelper.__angle_bracket_end, inline_request.next_index
        )
        new_token, between_brackets = None, None
        if closing_angle_index not in (-1, inline_request.next_index + 1):

            between_brackets, remaining_line = (
                inline_request.source_text[
                    inline_request.next_index + 1 : closing_angle_index
                ],
                inline_request.source_text[inline_request.next_index + 1 :],
            )
            closing_angle_index += 1

            new_column_number = inline_request.column_number + len(
                inline_request.remaining_line
            )

            new_token = InlineHelper.__parse_valid_uri_autolink(
                between_brackets, inline_request.line_number, new_column_number
            )
            if not new_token:
                new_token = InlineHelper.__parse_valid_email_autolink(
                    between_brackets, inline_request.line_number, new_column_number
                )
            if not new_token:
                new_token, after_index = HtmlHelper.parse_raw_html(
                    between_brackets,
                    remaining_line,
                    inline_request.line_number,
                    new_column_number,
                    inline_request,
                )
                if after_index != -1:
                    closing_angle_index = after_index + inline_request.next_index + 1
                    between_brackets = new_token.raw_tag

        inline_response = InlineResponse()
        if new_token:
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.new_tokens,
                between_brackets,
            ) = (
                "",
                closing_angle_index,
                [new_token],
                f"{InlineHelper.angle_bracket_start}{between_brackets}{InlineHelper.__angle_bracket_end}",
            )
        else:
            inline_response.new_string, inline_response.new_index, between_brackets = (
                InlineHelper.angle_bracket_start,
                inline_request.next_index + 1,
                InlineHelper.angle_bracket_start,
            )

        (
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ParserHelper.calculate_deltas(between_brackets)
        return inline_response



"""
Module to provide tests related to the MD026 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest


@pytest.mark.rules
def test_md025_bad_configuration_level():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.level=1",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md025.level' must be of type 'int'."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_good_configuration_level():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.level=$#1",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_configuration_level_bad():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.level=$#0",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md025.level' is not valid: Allowable values are between 1 and 6."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_configuration_front_matter_title():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.front_matter_title=$#1",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md025.front_matter_title' must be of type 'str'."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_good_configuration_front_matter_title():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.front_matter_title=subject",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_configuration_front_matter_title_bad():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.front_matter_title=",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md025.front_matter_title' is not valid: Empty strings are not allowable values."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_configuration_front_matter_title_invalid():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "plugins.md025.front_matter_title=a:b",
        "--strict-config",
        "scan",
        "test/resources/rules/md025/good_single_top_level.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while configuring plugins:\n"
        + "The value for property 'plugins.md025.front_matter_title' is not valid: Colons (:) are not allowed in the value."
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_good_single_top_level_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md025/good_single_top_level_atx.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_good_single_top_level_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md025/good_single_top_level_setext.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_top_level_atx_top_level_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md025/bad_top_level_atx_top_level_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md025/bad_top_level_atx_top_level_atx.md:5:1: "
        + "MD025: Multiple top-level headings in the same document (single-title,single-h1)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_top_level_atx_top_level_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md003",
        "scan",
        "test/resources/rules/md025/bad_top_level_atx_top_level_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md025/bad_top_level_atx_top_level_setext.md:6:1: "
        + "MD025: Multiple top-level headings in the same document (single-title,single-h1)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_top_level_setext_top_level_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md025/bad_top_level_setext_top_level_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md025/bad_top_level_setext_top_level_setext.md:7:1: "
        + "MD025: Multiple top-level headings in the same document (single-title,single-h1)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_top_level_setext_top_level_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md003",
        "scan",
        "test/resources/rules/md025/bad_top_level_setext_top_level_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md025/bad_top_level_setext_top_level_atx.md:6:1: "
        + "MD025: Multiple top-level headings in the same document (single-title,single-h1)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_good_front_matter_title():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "extensions.front-matter.enabled=$!True",
        "scan",
        "test/resources/rules/md025/good_front_matter_title.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_front_matter_title_top_level_atx():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "extensions.front-matter.enabled=$!True",
        "scan",
        "test/resources/rules/md025/bad_front_matter_title_top_level_atx.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md025/bad_front_matter_title_top_level_atx.md:7:1: "
        + "MD025: Multiple top-level headings in the same document (single-title,single-h1)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md025_bad_front_matter_title_top_level_setext():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md004 directory that has consistent asterisk usage on a single
    level list.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--set",
        "extensions.front-matter.enabled=$!True",
        "scan",
        "test/resources/rules/md025/bad_front_matter_title_top_level_setext.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md025/bad_front_matter_title_top_level_setext.md:8:1: "
        + "MD025: Multiple top-level headings in the same document (single-title,single-h1)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to handle the calculationof list looseness for the GRM transformer.
"""
import logging

from pymarkdown.parser_logger import ParserLogger

POGGER = ParserLogger(logging.getLogger(__name__))


class TransformToGfmListLooseness:
    """
    Class to handle the calculationof list looseness for the GRM transformer.
    """

    @staticmethod
    def calculate_list_looseness(actual_tokens, actual_token_index, next_token):
        """
        Based on the first token in a list, compute the "looseness" of the list.
        """

        POGGER.debug("\n\n__calculate_list_looseness>>$", actual_token_index)
        is_loose, current_token_index, stack_count = False, actual_token_index + 1, 0
        while True:
            check_me, stop_me, current_token = (
                False,
                False,
                actual_tokens[current_token_index],
            )
            if current_token.is_list_start:
                POGGER.debug("cll>>start list>>$", current_token)
                check_me, stack_count = TransformToGfmListLooseness.__handle_list_start(
                    stack_count
                )
            elif current_token.is_new_list_item:
                POGGER.debug("cll>>new list item>>$", current_token)
                check_me = TransformToGfmListLooseness.__handle_new_list_item(
                    current_token, stack_count
                )
            elif current_token.is_block_quote_start:
                POGGER.debug("cll>>start block quote>>$", current_token)
                stack_count = TransformToGfmListLooseness.__handle_block_quote_start(
                    stack_count
                )
            elif current_token.is_block_quote_end:
                POGGER.debug("cll>>end block quote>>$", current_token)
                stack_count = TransformToGfmListLooseness.__handle_block_quote_end(
                    stack_count
                )
            elif current_token.is_list_end:
                POGGER.debug("cll>>end list>>$", current_token)
                (
                    stop_me,
                    is_loose,
                    stack_count,
                ) = TransformToGfmListLooseness.__handle_list_end(
                    stack_count,
                    is_loose,
                    stop_me,
                    actual_tokens,
                    current_token_index,
                )
            elif actual_tokens[current_token_index - 1].is_blank_line:
                check_me = TransformToGfmListLooseness.__handle_blank_line(
                    current_token, stack_count, actual_tokens, current_token_index
                )

            POGGER.debug(
                ">>stack_count>>$>>#$:$>>check=$",
                stack_count,
                current_token_index,
                actual_tokens[current_token_index],
                check_me,
            )
            if check_me:
                POGGER.debug("check-->?")
                if TransformToGfmListLooseness.__is_token_loose(
                    actual_tokens, current_token_index
                ):
                    is_loose = True
                    stop_me = True
                    POGGER.debug("check-->Loose")
                else:
                    POGGER.debug("check-->Normal")
            if stop_me:
                break
            current_token_index += 1

        assert current_token_index != len(actual_tokens)
        POGGER.debug(
            "__calculate_list_looseness<<$<<$\n\n",
            actual_token_index,
            is_loose,
        )
        next_token.is_loose = is_loose
        return is_loose

    @staticmethod
    def __handle_list_start(stack_count):
        POGGER.debug(">>list--new>>$", stack_count)
        return stack_count == 0, stack_count + 1

    @staticmethod
    def __handle_new_list_item(current_token, stack_count):
        assert not current_token.is_block
        POGGER.debug(">>list--item>>$", stack_count)
        return stack_count == 0

    @staticmethod
    def __handle_list_end(
        stack_count,
        is_loose,
        stop_me,
        actual_tokens,
        current_token_index,
    ):
        if stack_count == 0:
            stop_me = True
        else:
            stack_count -= 1
            if TransformToGfmListLooseness.__correct_for_me(
                actual_tokens, current_token_index
            ):
                is_loose = True
                stop_me = True
                POGGER.debug("!!!latent-LOOSE!!!")
        POGGER.debug(">>list--end>>$", stack_count)
        return stop_me, is_loose, stack_count

    @staticmethod
    def __handle_block_quote_start(stack_count):
        POGGER.debug(">>block--new>>$", stack_count)
        return stack_count + 1

    @staticmethod
    def __handle_block_quote_end(stack_count):
        POGGER.debug(">>block--end>>$", stack_count)
        return stack_count - 1

    @staticmethod
    def __handle_blank_line(
        current_token, stack_count, actual_tokens, current_token_index
    ):
        search_back_index = current_token_index - 2
        pre_prev_token = actual_tokens[search_back_index]
        POGGER.debug(">>pre_prev_token>>$", pre_prev_token)

        while pre_prev_token.is_blank_line:
            search_back_index -= 1
            pre_prev_token = actual_tokens[search_back_index]

        if pre_prev_token.is_end_token:
            assert pre_prev_token.start_markdown_token, str(pre_prev_token)
            pre_prev_token = pre_prev_token.start_markdown_token
            POGGER.debug(">>end_>using_start>>$", pre_prev_token)

        current_check = (
            current_token.is_block and not current_token.is_link_reference_definition
        )
        pre_prev_check = (
            pre_prev_token.is_block and not pre_prev_token.is_link_reference_definition
        )

        POGGER.debug(">>other--stack_count>>$", stack_count)
        POGGER.debug(
            ">>other--current_token>>$>>$",
            current_token,
            current_check,
        )
        POGGER.debug(
            ">>other--current_token-2>>$>>$",
            pre_prev_token,
            pre_prev_check,
        )
        return stack_count == 0 and current_check and pre_prev_check

    @staticmethod
    def __correct_for_me(actual_tokens, current_token_index):
        correct_closure, is_valid = False, True
        assert current_token_index > 0

        POGGER.debug(">>prev>>$", actual_tokens[current_token_index - 1])
        if actual_tokens[current_token_index - 1].is_blank_line:
            search_index, actual_tokens_size = current_token_index + 1, len(
                actual_tokens
            )
            while (
                search_index < actual_tokens_size
                and actual_tokens[search_index].is_list_end
            ):
                search_index += 1
            POGGER.debug(">>ss>>$>>len>>$", search_index, actual_tokens_size)
            is_valid = search_index != actual_tokens_size
        if is_valid:
            POGGER.debug(">>current>>$", actual_tokens[current_token_index])
            POGGER.debug(">>current-1>>$", actual_tokens[current_token_index - 1])
            correct_closure = TransformToGfmListLooseness.__is_token_loose(
                actual_tokens, current_token_index
            )
            POGGER.debug(">>correct_closure>>$", correct_closure)
        return correct_closure

    @staticmethod
    def __is_token_loose(actual_tokens, current_token_index):
        """
        Check to see if this token inspires looseness.
        """

        check_index = current_token_index - 1
        token_to_check = actual_tokens[check_index]
        POGGER.debug("token_to_check-->$", token_to_check)

        while token_to_check.is_link_reference_definition:
            check_index -= 1
            token_to_check = actual_tokens[check_index]

        POGGER.debug("token_to_check-->$", token_to_check)
        if token_to_check.is_blank_line:
            POGGER.debug("before_blank-->$", actual_tokens[check_index - 1])
            if (
                actual_tokens[check_index - 1].is_new_list_item
                or actual_tokens[check_index - 1].is_list_start
            ):
                POGGER.debug("!!!Starting Blank!!!")
            else:
                POGGER.debug("!!!LOOSE!!!")
                return True
        return False

    @staticmethod
    def __find_owning_list_start(actual_tokens, actual_token_index):
        """
        Figure out what the list start for the current token is.
        """

        assert not actual_tokens[actual_token_index].is_list_start
        current_index, keep_going, stack_count = actual_token_index - 1, True, 0
        while keep_going and current_index >= 0:
            if actual_tokens[current_index].is_list_start:
                if stack_count == 0:
                    keep_going = False
                else:
                    stack_count -= 1
            elif actual_tokens[current_index].is_list_end:
                stack_count += 1
            if keep_going:
                current_index -= 1
        return current_index

    @staticmethod
    def reset_list_looseness(actual_tokens, actual_token_index):
        """
        Based on where we are within the actual tokens being emitted, figure
        out the correct list looseness to use.
        """

        POGGER.debug("!!!!!!!!!!!!!!!$", actual_token_index)
        search_index, stack_count, actual_tokens_size = (
            actual_token_index + 1,
            0,
            len(actual_tokens),
        )
        while search_index < actual_tokens_size:
            POGGER.debug(
                "!!$::$::$",
                stack_count,
                search_index,
                actual_tokens[search_index],
            )
            if actual_tokens[search_index].is_list_start:
                stack_count += 1
            elif actual_tokens[search_index].is_list_end:
                if not stack_count:
                    break
                stack_count -= 1
            search_index += 1
        POGGER.debug("!!!!!!!!!!!!!!!$-of-$", search_index, actual_tokens_size)
        # check to see where we are, then grab the matching start to find
        # the loose
        if search_index == actual_tokens_size:
            is_in_loose_list = True
        else:
            POGGER.debug(">>reset_list_looseness-token_list_start>>")
            new_index = TransformToGfmListLooseness.__find_owning_list_start(
                actual_tokens, search_index
            )
            POGGER.debug(">>reset_list_looseness>>$", new_index)
            is_in_loose_list = actual_tokens[new_index].is_loose
        POGGER.debug("           is_in_loose_list=$", is_in_loose_list)
        return is_in_loose_list



"""
Module to provide tests related to the plugin manager for the scanner.
"""
import os
from test.markdown_scanner import MarkdownScanner

from .utils import write_temporary_configuration

# pylint: disable=too-many-lines


def test_markdown_with_plugins_only():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "plugins",
    ]

    expected_return_code = 2
    expected_output = """usage: main.py plugins [-h] {list,info} ...

positional arguments:
  {list,info}
    list       list the available plugins
    info       information on a specific plugin

optional arguments:
  -h, --help   show this help message and exit
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_and_bad_path():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "MD047",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while loading plugins:\n"
        + "Plugin path 'MD047' does not exist.\n"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_and_single_plugin_file():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/plugin_two.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 0
    expected_output = """MD998>>init_from_config
MD998>>starting_new_file>>
MD998>>next_line:# This is a test
MD998>>next_line:
MD998>>next_line:The line after this line should be blank.
MD998>>next_line:
MD998>>completed_file
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_and_single_plugin_directory():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 0
    expected_output = """MD998>>init_from_config
MD998>>starting_new_file>>
MD998>>next_line:# This is a test
MD998>>next_line:
MD998>>next_line:The line after this line should be blank.
MD998>>next_line:
MD998>>completed_file
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_repeated_identifier():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/duplicate_id_debug.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """ValueError encountered while initializing plugins:
Unable to register plugin 'duplicate_id_debug.py' with id 'md999' as plugin 'plugin_one.py' is already registered with that id."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_bad_identifier():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_id.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """ValueError encountered while initializing plugins:
Unable to register plugin 'bad_id.py' with id 'debug-only' as id is not a valid id in the form 'aannn' or 'aaannn'."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_repeated_name():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/duplicate_name_debug.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """ValueError encountered while initializing plugins:
Unable to register plugin 'duplicate_name_debug.py' with name 'debug-only' as plugin 'plugin_one.py' is already registered with that name."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_bad_name():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_name.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """ValueError encountered while initializing plugins:
Unable to register plugin 'bad_name.py' with name 'debug.only' as name is not a valid name in the form 'an-an'."""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_and_bad_plugin_file():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/not-a-python-file.txt",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while loading plugins:\n"
        + "Plugin file named 'not-a-python-file.txt' cannot be loaded.\n"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_and_missing_class():
    """
    Test to make sure we get enable a rule if '--add-plugin' is supplied.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/misnamed.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while loading plugins:\n"
        + "Plugin file named 'misnamed.py' does not contain a class named 'Misnamed'.\n"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_starting_new_file():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the starting_new_file function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_starting_new_file.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
Plugin id 'MDE001' had a critical failure during the 'starting_new_file' action.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_completed_file():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the completed_file function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_completed_file.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
Plugin id 'MDE002' had a critical failure during the 'completed_file' action.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_next_line():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the next_line function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_next_line.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
(Line 1): Plugin id 'MDE003' had a critical failure during the 'next_line' action.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_next_line_with_stack_trace():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the next_line function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--stack-trace",
        "--add-plugin",
        "test/resources/plugins/bad/bad_next_line.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
(Line 1): Plugin id 'MDE003' had a critical failure during the 'next_line' action.
Actual Line: # This is a test
Traceback (most recent call last):
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output,
        expected_error,
        expected_return_code,
        additional_error=[
            """    raise Exception("bad next_line")
Exception: bad next_line

The above exception was the direct cause of the following exception:

Traceback (most recent call last):
""",
            """, in next_line
    raise BadPluginError(
pymarkdown.plugin_manager.BadPluginError: (Line 1): Plugin id 'MDE003' had a critical failure during the 'next_line' action.
""",
        ],
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_next_line_with_configuration_stack_trace():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the next_line function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--stack-trace",
        "--add-plugin",
        "test/resources/plugins/bad/bad_next_line.py",
        "--set",
        "log.stack-trace=$!True",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
(Line 1): Plugin id 'MDE003' had a critical failure during the 'next_line' action.
Actual Line: # This is a test
Traceback (most recent call last):
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output,
        expected_error,
        expected_return_code,
        additional_error=[
            """    raise Exception("bad next_line")
Exception: bad next_line

The above exception was the direct cause of the following exception:

Traceback (most recent call last):
""",
            """, in next_line
    raise BadPluginError(
pymarkdown.plugin_manager.BadPluginError: (Line 1): Plugin id 'MDE003' had a critical failure during the 'next_line' action.
""",
        ],
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_next_token():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the next_token function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_next_token.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
(1,1): Plugin id 'MDE003' had a critical failure during the 'next_token' action.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_next_token_with_stack_trace():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the next_token function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--stack-trace",
        "--add-plugin",
        "test/resources/plugins/bad/bad_next_token.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while scanning 'test/resources/rules/md047/end_with_blank_line.md':
(1,1): Plugin id 'MDE003' had a critical failure during the 'next_token' action.
Actual Token: [atx(1,1):1:0:]
Traceback (most recent call last):
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output,
        expected_error,
        expected_return_code,
        additional_error=[
            """    raise Exception("bad next_token")
Exception: bad next_token

The above exception was the direct cause of the following exception:

Traceback (most recent call last):
""",
            """, in next_token
    raise BadPluginError(
pymarkdown.plugin_manager.BadPluginError: (1,1): Plugin id 'MDE003' had a critical failure during the 'next_token' action.
""",
        ],
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_constructor():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the constructor function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_constructor.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while loading plugins:\n"
        + "Plugin file named 'bad_constructor.py' threw an exception in the constructor for the class 'BadConstructor'.\n"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_details():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the details function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_details.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadDetails' had a critical failure loading the plugin details.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_details_with_stack_trace():
    """
    Test to make sure we get an error logged if a plugin throws an exception within the details function.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--stack-trace",
        "--add-plugin",
        "test/resources/plugins/bad/bad_details.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadDetails' had a critical failure loading the plugin details.
Traceback (most recent call last):
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output,
        expected_error,
        expected_return_code,
        additional_error=[
            """    raise Exception("bad details")
Exception: bad details

The above exception was the direct cause of the following exception:

Traceback (most recent call last):
""",
            """    raise BadPluginError(
pymarkdown.plugin_manager.BadPluginError: Plugin class 'BadDetails' had a critical failure loading the plugin details.""",
        ],
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_string_detail():
    """
    Test to make sure we get an error logged if a plugin throws an exception that a string detail is bad.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_string_detail_is_int.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadStringDetailIsInt' returned an improperly typed value for field name 'plugin_description'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_string_detail_from_configuration():
    """
    Test to make sure we get an error logged if a plugin throws an exception that a string detail is bad.
    Note: this version loads from configuration.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_configuration = {
        "plugins": {
            "additional_paths": "test/resources/plugins/bad/bad_string_detail_is_int.py"
        }
    }
    configuration_file = None
    try:
        configuration_file = write_temporary_configuration(supplied_configuration)
        supplied_arguments = [
            "-c",
            configuration_file,
            "scan",
            "test/resources/rules/md047/end_with_blank_line.md",
        ]

        expected_return_code = 1
        expected_output = ""
        expected_error = """\n\nBadPluginError encountered while loading plugins:
Plugin class 'BadStringDetailIsInt' returned an improperly typed value for field name 'plugin_description'.
"""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
    finally:
        if configuration_file and os.path.exists(configuration_file):
            os.remove(configuration_file)


def test_markdown_with_dash_dash_add_plugin_with_empty_string_detail():
    """
    Test to make sure we get an error logged if a plugin throws an exception that a string detail is empty.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_string_detail_is_empty.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadStringDetailIsEmpty' returned an empty value for field name 'plugin_description'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_boolean_detail():
    """
    Test to make sure we get an error logged if a plugin throws an exception that a boolean detail is bad.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_boolean_detail_is_int.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadBooleanDetailIsInt' returned an improperly typed value for field name 'plugin_enabled_by_default'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_integer_detail():
    """
    Test to make sure we get an error logged if a plugin throws an exception that an integer detail is bad.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_integer_detail_is_string.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadIntegerDetailIsString' returned an improperly typed value for field name 'plugin_interface_version'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_description():
    """
    Test to make sure we get an error logged if a plugin returns a bad description.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_description.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'BadDescription' returned an improperly typed value for field name 'plugin_description'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_empty_description():
    """
    Test to make sure we get an error logged if a plugin returns an empty description.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/empty_description.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin class 'EmptyDescription' returned an empty value for field name 'plugin_description'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_blank_description():
    """
    Test to make sure we get an error logged if a plugin returns a blank description.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/blank_description.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """ValueError encountered while initializing plugins:
Unable to register plugin 'blank_description.py' with a description string that is blank.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_semantic_version():
    """
    Test to make sure we get an error logged if a plugin returns a bad semantic version.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_semantic_version.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """ValueError encountered while initializing plugins:
Unable to register plugin 'bad_semantic_version.py' with a version string that is not a valid semantic version.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_dash_dash_add_plugin_with_bad_interface_version():
    """
    Test to make sure we get an error logged if a plugin returns a bad interface version.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--add-plugin",
        "test/resources/plugins/bad/bad_interface_version.py",
        "scan",
        "test/resources/rules/md047/end_with_blank_line.md",
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = """BadPluginError encountered while loading plugins:
Plugin 'bad_interface_version.py' with an interface version ('2') that is not '1'.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_list_only():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "list"]

    expected_return_code = 0
    expected_output = """
  ID     NAMES                                   ENABLED    ENABLED    VERSION
                                                 (DEFAULT)  (CURRENT)

  md001  heading-increment, header-increment     True       True       0.5.0
  md002  first-heading-h1, first-header-h1       False      False      0.5.0
  md003  heading-style, header-style             True       True       0.5.0
  md004  ul-style                                True       True       0.5.0
  md005  list-indent                             True       True       0.5.0
  md006  ul-start-left                           False      False      0.5.0
  md007  ul-indent                               True       True       0.5.0
  md009  no-trailing-spaces                      True       True       0.5.0
  md010  no-hard-tabs                            True       True       0.5.0
  md011  no-reversed-links                       True       True       0.5.0
  md012  no-multiple-blanks                      True       True       0.5.0
  md013  line-length                             True       True       0.5.0
  md014  commands-show-output                    True       True       0.5.0
  md018  no-missing-space-atx                    True       True       0.5.0
  md019  no-multiple-space-atx                   True       True       0.5.0
  md020  no-missing-space-closed-atx             True       True       0.5.0
  md021  no-multiple-space-closed-atx            True       True       0.5.0
  md022  blanks-around-headings, blanks-around-  True       True       0.5.0
         headers
  md023  heading-start-left, header-start-left   True       True       0.5.0
  md024  no-duplicate-heading, no-duplicate-hea  True       True       0.5.0
         der
  md025  single-title, single-h1                 True       True       0.5.0
  md026  no-trailing-punctuation                 True       True       0.5.0
  md027  no-multiple-space-blockquote            True       True       0.5.0
  md028  no-blanks-blockquote                    True       True       0.5.0
  md029  ol-prefix                               True       True       0.5.0
  md030  list-marker-space                       True       True       0.5.0
  md031  blanks-around-fences                    True       True       0.5.0
  md032  blanks-around-lists                     True       True       0.5.0
  md033  no-inline-html                          True       True       0.5.0
  md034  no-bare-urls                            True       True       0.5.0
  md035  hr-style                                True       True       0.5.0
  md036  no-emphasis-as-heading, no-emphasis-as  True       True       0.5.0
         -header
  md037  no-space-in-emphasis                    True       True       0.5.0
  md038  no-space-in-code                        True       True       0.5.0
  md039  no-space-in-links                       True       True       0.5.0
  md040  fenced-code-language                    True       True       0.5.0
  md041  first-line-heading, first-line-h1       True       True       0.5.0
  md042  no-empty-links                          True       True       0.5.0
  md043  required-headings, required-headers     True       True       0.5.0
  md044  proper-names                            True       True       0.5.0
  md045  no-alt-text                             True       True       0.5.0
  md046  code-block-style                        True       True       0.5.0
  md047  single-trailing-newline                 True       True       0.5.0
  md048  code-fence-style                        True       True       0.5.0

"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, suppress_first_line_heading_rule=False
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_list_only_all():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "list", "--all"]

    expected_return_code = 0
    expected_output = """
  ID     NAMES                                   ENABLED    ENABLED    VERSION
                                                 (DEFAULT)  (CURRENT)

  md001  heading-increment, header-increment     True       True       0.5.0
  md002  first-heading-h1, first-header-h1       False      False      0.5.0
  md003  heading-style, header-style             True       True       0.5.0
  md004  ul-style                                True       True       0.5.0
  md005  list-indent                             True       True       0.5.0
  md006  ul-start-left                           False      False      0.5.0
  md007  ul-indent                               True       True       0.5.0
  md009  no-trailing-spaces                      True       True       0.5.0
  md010  no-hard-tabs                            True       True       0.5.0
  md011  no-reversed-links                       True       True       0.5.0
  md012  no-multiple-blanks                      True       True       0.5.0
  md013  line-length                             True       True       0.5.0
  md014  commands-show-output                    True       True       0.5.0
  md018  no-missing-space-atx                    True       True       0.5.0
  md019  no-multiple-space-atx                   True       True       0.5.0
  md020  no-missing-space-closed-atx             True       True       0.5.0
  md021  no-multiple-space-closed-atx            True       True       0.5.0
  md022  blanks-around-headings, blanks-around-  True       True       0.5.0
         headers
  md023  heading-start-left, header-start-left   True       True       0.5.0
  md024  no-duplicate-heading, no-duplicate-hea  True       True       0.5.0
         der
  md025  single-title, single-h1                 True       True       0.5.0
  md026  no-trailing-punctuation                 True       True       0.5.0
  md027  no-multiple-space-blockquote            True       True       0.5.0
  md028  no-blanks-blockquote                    True       True       0.5.0
  md029  ol-prefix                               True       True       0.5.0
  md030  list-marker-space                       True       True       0.5.0
  md031  blanks-around-fences                    True       True       0.5.0
  md032  blanks-around-lists                     True       True       0.5.0
  md033  no-inline-html                          True       True       0.5.0
  md034  no-bare-urls                            True       True       0.5.0
  md035  hr-style                                True       True       0.5.0
  md036  no-emphasis-as-heading, no-emphasis-as  True       True       0.5.0
         -header
  md037  no-space-in-emphasis                    True       True       0.5.0
  md038  no-space-in-code                        True       True       0.5.0
  md039  no-space-in-links                       True       True       0.5.0
  md040  fenced-code-language                    True       True       0.5.0
  md041  first-line-heading, first-line-h1       True       True       0.5.0
  md042  no-empty-links                          True       True       0.5.0
  md043  required-headings, required-headers     True       True       0.5.0
  md044  proper-names                            True       True       0.5.0
  md045  no-alt-text                             True       True       0.5.0
  md046  code-block-style                        True       True       0.5.0
  md047  single-trailing-newline                 True       True       0.5.0
  md048  code-fence-style                        True       True       0.5.0
  md999  debug-only                              False      False      0.0.0

"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, suppress_first_line_heading_rule=False
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_list_and_filter_by_id_ends_with_nine():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "list", "md*9"]

    expected_return_code = 0
    expected_output = """
  ID     NAMES                  ENABLED    ENABLED    VERSION
                                (DEFAULT)  (CURRENT)

  md009  no-trailing-spaces     True       True       0.5.0
  md019  no-multiple-space-atx  True       True       0.5.0
  md029  ol-prefix              True       True       0.5.0
  md039  no-space-in-links      True       True       0.5.0

"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_list_and_filter_by_id_ends_with_non_sequence():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "list", "this-is-not-an-used-identifier"]

    expected_return_code = 0
    expected_output = """No plugin rule identifiers matches the pattern 'this-is-not-an-used-identifier'."""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_list_and_filter_by_name_link():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "list", "*header*"]

    expected_return_code = 0
    expected_output = """
  ID     NAMES                                   ENABLED    ENABLED    VERSION
                                                 (DEFAULT)  (CURRENT)

  md001  heading-increment, header-increment     True       True       0.5.0
  md002  first-heading-h1, first-header-h1       False      False      0.5.0
  md003  heading-style, header-style             True       True       0.5.0
  md022  blanks-around-headings, blanks-around-  True       True       0.5.0
         headers
  md023  heading-start-left, header-start-left   True       True       0.5.0
  md024  no-duplicate-heading, no-duplicate-hea  True       True       0.5.0
         der
  md036  no-emphasis-as-heading, no-emphasis-as  True       True       0.5.0
         -header
  md043  required-headings, required-headers     True       True       0.5.0
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_list_and_bad_filter():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "list", "*"]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py plugins list [-h] [--all] [list_filter]
main.py plugins list: error: argument list_filter: Value '*' is not a valid pattern for an id or a name.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_info_and_bad_filter():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "info", "abc.def"]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py plugins info [-h] info_filter
main.py plugins info: error: argument info_filter: Value 'abc.def' is not a valid id or name.
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_info_and_not_found_filter():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "info", "md00001"]

    expected_return_code = 1
    expected_output = "Unable to find a plugin with an id or name of 'md00001'."
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_info_and_found_filter():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "info", "md001"]

    expected_return_code = 0
    expected_output = """  ITEM                 DESCRIPTION

  Id                   md001
  Name(s)              heading-increment,header-increment
  Short Description    Heading levels should only increment by one level at a
                       time.
  Description Url      https://github.com/jackdewinter/pymarkdown/blob/main/do
                       cs/rules/rule_md001.md
  Configuration Items  front_matter_title
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_info_and_found_filter_no_configuration():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "info", "md023"]

    expected_return_code = 0
    expected_output = """  ITEM               DESCRIPTION

  Id                 md023
  Name(s)            heading-start-left,header-start-left
  Short Description  Headings must start at the beginning of the line.
  Description Url    https://github.com/jackdewinter/pymarkdown/blob/main/docs
                     /rules/rule_md023.md
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_with_plugins_info_and_found_filter_no_configuration_and_no_url():
    """
    Test to make sure
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["plugins", "info", "md999"]

    expected_return_code = 0
    expected_output = """  ITEM               DESCRIPTION

  Id                 md999
  Name(s)            debug-only
  Short Description  Debug plugin
"""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
https://github.github.com/gfm/#task-list-items-extension-
"""

from .utils import act_and_assert


def test_task_list_items_279():
    """
    Test case 279:  If the character between the brackets is a whitespace character, the checkbox is unchecked. Otherwise, the checkbox is checked.
    """

    # Arrange
    source_markdown = """- [ ] foo
- [x] bar"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):[:]",
        "[text(1,4): :]",
        "[text(1,5):]:]",
        "[text(1,6): foo:]",
        "[end-para:::True]",
        "[li(2,1):2::]",
        "[para(2,3):]",
        "[text(2,3):[:]",
        "[text(2,4):x:]",
        "[text(2,5):]:]",
        "[text(2,6): bar:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>\n<li>[ ] foo</li>\n<li>[x] bar</li>\n</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


def test_task_list_items_280():
    """
    Test case 280:  Task lists can be arbitrarily nested:
    """

    # Arrange
    source_markdown = """- [x] foo
  - [ ] bar
  - [x] baz
- [ ] bim"""
    expected_tokens = [
        "[ulist(1,1):-::2:]",
        "[para(1,3):]",
        "[text(1,3):[:]",
        "[text(1,4):x:]",
        "[text(1,5):]:]",
        "[text(1,6): foo:]",
        "[end-para:::True]",
        "[ulist(2,3):-::4:  ]",
        "[para(2,5):]",
        "[text(2,5):[:]",
        "[text(2,6): :]",
        "[text(2,7):]:]",
        "[text(2,8): bar:]",
        "[end-para:::True]",
        "[li(3,3):4:  :]",
        "[para(3,5):]",
        "[text(3,5):[:]",
        "[text(3,6):x:]",
        "[text(3,7):]:]",
        "[text(3,8): baz:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
        "[li(4,1):2::]",
        "[para(4,3):]",
        "[text(4,3):[:]",
        "[text(4,4): :]",
        "[text(4,5):]:]",
        "[text(4,6): bim:]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>\n<li>[x] foo\n<ul>\n<li>[ ] bar</li>\n<li>[x] baz</li>\n</ul>\n</li>\n<li>[ ] bim</li>\n</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Tests for the collect_until_one_of_characters function.
"""
from pymarkdown.parser_helper import ParserHelper


def test_empty_string_with_good_index():
    """
    Make sure that an empty string is handled properly with a good index
    """

    # Arrange
    input_string = ""
    start_index = 0
    characters_to_match = " !"
    expected_output = (0, "")

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output


def test_empty_string_with_bad_right_index():
    """
    Make sure that an empty string is handled properly with an index that is too far to the right.
    """

    # Arrange
    input_string = ""
    start_index = 2
    characters_to_match = " !"
    expected_output = (None, None)

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output


def test_empty_string_with_bad_left_index():
    """
    Make sure that an empty string is handled properly with an index that is too far to the left.
    """

    # Arrange
    input_string = ""
    start_index = -1
    characters_to_match = " !"
    expected_output = (None, None)

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output


def test_simple_case_from_start():
    """
    Make sure that we test a simple extraction from the start of the string.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 0
    characters_to_match = " !"
    expected_output = (4, "this")

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output


def test_simple_case_from_middle():
    """
    Make sure that we test a simple extraction from the middle of the string.
    """

    # Arrange
    input_string = "this!is!a!test"
    start_index = 5
    characters_to_match = " !"
    expected_output = (7, "is")

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output


def test_simple_case_from_end():
    """
    Make sure that we test a simple extraction from the end of the string.
    """

    # Arrange
    input_string = "this is a test"
    start_index = 10
    characters_to_match = " !"
    expected_output = (len(input_string), "test")

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output


def test_already_on_whitespace():
    """
    Make sure that we test extracting while already on a whitespace character.
    """

    # Arrange
    input_string = "this!is!a!test"
    start_index = 9
    characters_to_match = " !"
    expected_output = (9, "")

    # Act
    actual_output = ParserHelper.collect_until_one_of_characters(
        input_string, start_index, characters_to_match
    )

    # Assert
    assert expected_output == actual_output
